//
// Created by illyum on 9/14/2024.
//

#pragma once

#include <pch.hpp>

namespace ICEngine {
	enum LogLevel {
		OFF = 0,
		TRACE,
		DEBUG,
		WARN,
		ERROR,
		CRITICAL,
		FATAL
	};

	class LogSink {
	public:
		virtual ~LogSink() = default;

		virtual void Log(const std::string &message) = 0;
	};

	class ConsoleSink : public LogSink {
	public:
		void Log(const std::string &message) override {
			std::cout << message << std::endl;
		}
	};

	class FileSink : public LogSink {
	public:
		explicit FileSink(const std::string &filename) {
			fileStream.open(filename);
		}

		void Log(const std::string &message) override {
			if (fileStream.is_open()) {
				fileStream << message << std::endl;
			}
		}

	private:
		std::ofstream fileStream;
	};

	class Logger {
	public:
		explicit Logger(const char *name) : name(name) {
		}

		void SetLevel(LogLevel level) {
			this->level = level;
		}

		void AddSink(std::shared_ptr<LogSink> sink) {
			sinks.push_back(std::move(sink));
		}

		template<typename... Args>
		void Log(LogLevel level, const char *file, int line, const char *func, const std::string &formatStr,
		         Args &&... args) {
			if (level < this->level) return;

			std::string logMessage = formatMessage(level, file, line, func,
			                                       format(formatStr, std::forward<Args>(args)...));

			for (const auto &sink: sinks) {
				sink->Log(logMessage);
			}
		}

	private:
		const char *name;
		std::vector<std::shared_ptr<LogSink> > sinks;
		LogLevel level = DEBUG;

		static const char *levelToString(LogLevel level) {
			switch (level) {
				case TRACE: return "TRACE";
				case DEBUG: return "DEBUG";
				case WARN: return "WARN";
				case ERROR: return "ERROR";
				case CRITICAL: return "CRITICAL";
				case FATAL: return "FATAL";
				default: return "UNKNOWN";
			}
		}

		static std::string getCurrentTime() {
			auto now = std::chrono::system_clock::now();
			std::time_t time = std::chrono::system_clock::to_time_t(now);
			std::ostringstream oss;
			oss << std::put_time(std::localtime(&time), "%Y/%m/%d %H:%M:%S");
			return oss.str();
		}

		static std::string truncateFilePath(const char *file) {
			std::string path(file);
			size_t pos = path.find_last_of("/\\");
			if (pos != std::string::npos && pos > 3) {
				return "..." + path.substr(pos - 3);
			}
			return path;
		}

		template<typename... Args>
		std::string format(const std::string &formatStr, Args &&... args) {
			std::ostringstream oss;
			formatHelper(oss, formatStr, std::forward<Args>(args)...);
			return oss.str();
		}

		void formatHelper(std::ostringstream &oss, const std::string &formatStr) {
			oss << formatStr;
		}

		template<typename T, typename... Args>
		void formatHelper(std::ostringstream &oss, const std::string &formatStr, T &&firstArg,
		                  Args &&... remainingArgs) {
			size_t pos = formatStr.find("{}");
			if (pos != std::string::npos) {
				oss << formatStr.substr(0, pos) << std::forward<T>(firstArg);
				formatHelper(oss, formatStr.substr(pos + 2), std::forward<Args>(remainingArgs)...);
			} else {
				oss << formatStr;
			}
		}

		std::string formatMessage(LogLevel level, const char *file, int line, const char *func,
		                          const std::string &message) {
			std::ostringstream oss;
			oss << getCurrentTime() << " ";
			oss << truncateFilePath(file) << ":" << line << ":" << func << " ";
			oss << "[" << name << "] ";
			oss << "[" << levelToString(level) << "]: ";
			oss << message;
			return oss.str();
		}
	};

	class Log {
	public:
		static void Init() {
			static bool isInitialized = false;
			if (isInitialized) {
				GetCoreLogger()->Log(WARN, __FILE__, __LINE__, __FUNCTION__,
				                     "Logger already initialized. Skipping re-initialization.");
				return;
			}
			GetCoreLogger();
			GetAppLogger();
			isInitialized = true;
		}

		static std::shared_ptr<Logger> &GetCoreLogger() {
			static std::shared_ptr<Logger> CoreLogger(new Logger("CORE"));
			return CoreLogger;
		}

		static std::shared_ptr<Logger> &GetAppLogger() {
			static std::shared_ptr<Logger> AppLogger(new Logger("APP"));
			return AppLogger;
		}

	private:
		Log() = delete;

		Log(const Log &) = delete;

		Log &operator=(const Log &) = delete;
	};

#define CORE_LOG_FATAL(message) \
	do { \
	ICEngine::Log::GetCoreLogger()->Log(ICEngine::FATAL, __FILE__, __LINE__, __FUNCTION__, message); \
	ICEngine::ICEApplication::HandleFatalError(message); \
	} while (0)

#define LOG_FATAL(message) \
	do { \
	ICEngine::Log::GetAppLogger()->Log(ICEngine::FATAL, __FILE__, __LINE__, __FUNCTION__, message); \
	ICEngine::ICEApplication::HandleFatalError(message); \
	} while (0)

#ifndef NDEBUG
    #define CORE_LOG_TRACE(message, ...)    ICEngine::Log::GetCoreLogger()->Log(ICEngine::TRACE, __FILE__, __LINE__, __FUNCTION__, message, ##__VA_ARGS__)
    #define CORE_LOG_DEBUG(message, ...)    ICEngine::Log::GetCoreLogger()->Log(ICEngine::DEBUG, __FILE__, __LINE__, __FUNCTION__, message, ##__VA_ARGS__)
    #define CORE_LOG_WARN(message, ...)     ICEngine::Log::GetCoreLogger()->Log(ICEngine::WARN, __FILE__, __LINE__, __FUNCTION__, message, ##__VA_ARGS__)
    #define CORE_LOG_ERROR(message, ...)    ICEngine::Log::GetCoreLogger()->Log(ICEngine::ERROR, __FILE__, __LINE__, __FUNCTION__, message, ##__VA_ARGS__)
    #define CORE_LOG_CRITICAL(message, ...) ICEngine::Log::GetCoreLogger()->Log(ICEngine::CRITICAL, __FILE__, __LINE__, __FUNCTION__, message, ##__VA_ARGS__)

    #define LOG_TRACE(message, ...)         ICEngine::Log::GetAppLogger()->Log(ICEngine::TRACE, __FILE__, __LINE__, __FUNCTION__, message, ##__VA_ARGS__)
    #define LOG_DEBUG(message, ...)         ICEngine::Log::GetAppLogger()->Log(ICEngine::DEBUG, __FILE__, __LINE__, __FUNCTION__, message, ##__VA_ARGS__)
    #define LOG_WARN(message, ...)          ICEngine::Log::GetAppLogger()->Log(ICEngine::WARN, __FILE__, __LINE__, __FUNCTION__, message, ##__VA_ARGS__)
    #define LOG_ERROR(message, ...)         ICEngine::Log::GetAppLogger()->Log(ICEngine::ERROR, __FILE__, __LINE__, __FUNCTION__, message, ##__VA_ARGS__)
    #define LOG_CRITICAL(message, ...)      ICEngine::Log::GetAppLogger()->Log(ICEngine::CRITICAL, __FILE__, __LINE__, __FUNCTION__, message, ##__VA_ARGS__)
#else
    #define CORE_LOG_TRACE(message, ...)
    #define CORE_LOG_DEBUG(message, ...)
    #define CORE_LOG_WARN(message, ...)
    #define CORE_LOG_ERROR(message, ...)
    #define CORE_LOG_CRITICAL(message, ...)

    #define LOG_TRACE(message, ...)
    #define LOG_DEBUG(message, ...)
    #define LOG_WARN(message, ...)
    #define LOG_ERROR(message, ...)
    #define LOG_CRITICAL(message, ...)
#endif
}
